// mozileUndoRedo.js

/* ***** BEGIN LICENSE BLOCK *****
 * Licensed under Version: MPL 1.1/GPL 2.0/LGPL 2.1
 * Full Terms at http://mozile.mozdev.org/license.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Playsophy code.
 *
 * The Initial Developer of the Original Code is Playsophy
 * Portions created by the Initial Developer are Copyright (C) 2002-2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *		Karl Guertin <grayrest@grayrest.com>
 *		James A. Overton <james@overton.ca>
 *		Christian Stocker (Bitflux)
 *		Paul Everitt
 *		David Palm
 *
 * ***** END LICENSE BLOCK ***** */

/*******

Mozile Undo Redo Javascript
Mozile version 0.61

Created by James A. Overton
Created on 2004-03-24

Last revised by James A. Overton
Last revised on 2004-04-17

HISTORY:
	Some code in this file was first found in mozilehtmltb.js, while the 
majority was cut out of mozileTB.js and separated into different files.


DESCRIPTION:
	This code handles the mozileUndo and mozileRedo functions, as well as the
mozileDo function.

*********/




/********** Undo Functionality *************/

// This is the lazy way, I copy the whole BODY...
// This code still needs to be tweaked.


// Mozile Do
// Records the state of the current document into the undoStack for that document,
// and increments the undoCounter (each document gets its own stack and counter).
// All steps higher up in the stack are eliminated when a lower one is added.
// If the undoMax number of steps is hit, then the stack is spliced.
// The undoStack includes the name of the command and the whole BODY node.

// Yes, it is a brute force solution to copy the whole BODY,
// but it is the best solution for multiple contentEditable areas,
// (doing the whole HTML messes up the scrollbar)
// and the performance seems fine.
// Of course this won't work for XML in general...

const MOZILE_UNDO_NAME = 0;
const MOZILE_UNDO_NODE = 1;
const MOZILE_UNDO_FOCUS_NODE = 2;
const MOZILE_UNDO_FOCUS_OFFSET = 3;

//Initialisation required to add the undoStack etc. to the current target.
function mozileUndoInit() {
	if( !("undoStack" in mozileTarget) ) {
		mozileTarget.undoStack=new Array();
		mozileTarget.undoCounter=0;
	}
}

function mozileDo(name) {
	var f = ["mozileUndoRedo.js","mozileDo"];
	mozileDebug(f,3,"Do: "+name);
	
	try{
		//Initilisation. (Not the best place because it should only need to be called once.)
		mozileUndoInit();
		
		//Prepare the data for the step to be added to the undo stack.
		var undoStep = new Array;
		undoStep[MOZILE_UNDO_NAME]=name;	//Store step name.
		
		//Get the last detected element node and index.
		var ename, enamespace, eoffset, elist;
		if ( (mozile.lastContextNode) && (mozile.lastContextNode.nodeType == Node.ELEMENT_NODE) ) {
			ename = mozile.lastContextNode.localName;
			enamespace = mozile.lastContextNode.namespaceURI;
			elist = mozileTarget.document.getElementsByTagNameNS( enamespace, ename );
			for ( var i=0; i<elist.length; i++ ) {
				if ( elist[i] === mozile.lastContextNode ) {
					eoffset = i;
					break;
				}
			}
		} else {
			mozileDebug(f,0,"No last context node");
		}
		//Store Element name and index.
		if ( (ename != undefined) && (eoffset != undefined) ) {
			mozileDebug(f,3,"namespace: "+enamespace+" /name: "+ename+" /offset: "+eoffset+ " /elist.length: "+elist.length);
			undoStep[MOZILE_UNDO_FOCUS_NODE] = ename;
			undoStep[MOZILE_UNDO_FOCUS_OFFSET] = eoffset;
		} else {
			mozileDebug(f,0,"No name/offset");
			undoStep[MOZILE_UNDO_FOCUS_NODE] = null;
			undoStep[MOZILE_UNDO_FOCUS_OFFSET] = null;
		}

		var getNode;
		//If x/html document clone the whole <body>...</body>
		if(mozileTarget.document.documentElement.tagName.toLowerCase()=="html") {
			mozileDebug(f,3,"X/HTML context: "+ mozileTarget.document.documentElement.tagName);
			getNode = mozileTarget.document.getElementsByTagName("body")[0];
		}
		else { //If not an x/html document clone the whole document element.
			mozileDebug(f,3,"XML context: "+ mozileTarget.document.documentElement.tagName);
			getNode=mozileTarget.document.documentElement;
		}
		undoStep[MOZILE_UNDO_NODE]=getNode.cloneNode(true);

		//undoStep['position']=getPosition(getNode);
		mozileTarget.undoStack=mozileTarget.undoStack.slice(0,mozileTarget.undoCounter);			// make sure there are no extra steps
		if(mozileTarget.undoStack.length > undoMax) {
			mozileTarget.undoStack=mozileTarget.undoStack.slice(1);			// make sure there are no extra steps
		}
		mozileTarget.undoStack.push(undoStep);
		mozileTarget.undoCounter=mozileTarget.undoStack.length;
		//alert("Undo:"+mozileTarget.undoStack.length +":"+mozileTarget.undoCounter);
		mozileUpdateUndoButtons();
		mozileTarget.document.MOZILE_CURRENT_CONFIG.saved = "false";
	} catch(e) {
		mozileDebug(f,1,"Error with the document:" + e);
	}
} // end mozileDo


// Update Undo Buttons
// It gives the user some feedback about when more undo/redo steps are available.
//
function updateUndoButtons() {
	var f = ["mozileUndoRedo.js","updateUndoButtons"];
	var u=document.getElementById('mozile-undo');
	var r=document.getElementById('mozile-redo');
	try {
  	if(mozileTarget.undoCounter<1) {
  		u.setAttribute('disabled','true');
  	}else {
  		u.removeAttribute('disabled');
  	}
  	if(mozileTarget.undoCounter>mozileTarget.undoStack.length-2) {
  		r.setAttribute('disabled','true');
  	}else {
  		r.removeAttribute('disabled');
  	}
	}
	catch (e) {
		mozileDebug(f,1,"Exception! "+e);
	}
} // end updateUndoButtons()


// Mozile Undo
// Moves down the undoStack from the curent undoCounter,
// replacing the current BODY node with the one from the stack.
// Behaves a bit strangely around the undoMax limit.

function mozileUndo() {
	var f = ["mozileUndoRedo.js","mozileUndo"];
	try {
  	if( !("undoCounter" in mozileTarget) || mozileTarget.undoCounter<1) {
  		return; //Nothing to undo.
  	}
  
  	var undoStack=mozileTarget.undoStack;				// it is safe to copy this, since it will not be manipulated
  	
  	if(mozileTarget.undoCounter==undoStack.length) {
  		mozileDo('Current');
   		mozileTarget.undoCounter--;
  	}
   	mozileTarget.undoCounter--;
  	var undoStep=undoStack[mozileTarget.undoCounter];
  
  	mozileReplaceNode(undoStep[MOZILE_UNDO_NODE]);
  	mozileRestoreFocus( undoStep );
  	mozileUpdateUndoButtons();
	} catch (e) {
		mozileDebug(f,0,"Exception: "+e);
	}
} // end mozileUndo()

// Mozile Redo
// Moves up the undoStack from the current undoCounter,
// replacing the current BODY node with the one in the stack.

function mozileRedo() {
	var f = ["mozileUndoRedo.js","mozileRedo"];
	try {
  	if ( !("undoCounter" in mozileTarget) )
  		return;

  	var undoStack=mozileTarget.undoStack;				// it is safe to copy this, since it will not be manipulated
  	if(mozileTarget.undoCounter>undoStack.length-2)
  		return;
  
   	mozileTarget.undoCounter++;
  	var undoStep=undoStack[mozileTarget.undoCounter];
  	mozileReplaceNode(undoStep[MOZILE_UNDO_NODE]);
  	mozileRestoreFocus( undoStep );
  	mozileUpdateUndoButtons();
	} catch (e) {
		mozileDebug(f,0,"Exception: "+e);
	}
} // end mozileRedo()


//Finds the element in the document and restores the focus to the start of the element.
//	
function mozileRestoreFocus( undoStep ) {
	var f = ["mozileUndoRedo.js","mozileRestoreFocus"];
	try {
  	var fnode = undoStep[MOZILE_UNDO_FOCUS_NODE];
  	var foffset = undoStep[MOZILE_UNDO_FOCUS_OFFSET];
  	if ( (fnode != null) && (foffset != null) ) {
			var elist = mozileTarget.document.getElementsByTagName( fnode );
  		if ( foffset < elist.length ) {
  			var currentSelection = mozileTarget.getSelection();
  			currentSelection.collapse( elist[foffset], 0 );
  		}
  	} else {
			mozileDebug(f,0,"No focus to restore!");
		}
	} catch (e) {
		mozileDebug(f,0,"Exception: "+e);
	}
}	


// Replace a special target node with the replacement.
// The target will depend on the content, for HTML it is the body tag,
// for XML it is currently the document element.
//
function mozileReplaceNode(replacement) {
	var f = ["mozileUndoRedo.js","mozileReplaceNode"];	
	mozileDebug(f,3,"Replacing Node "+ replacement);
	var target, rChildren, rLength, replacementFragment, i;
	
	if(mozileTarget.document.documentElement.tagName.toLowerCase()=="html") {
		mozileDebug(f,3,"X/HTML context");
		try {
			replacementFragment = mozileTarget.document.createDocumentFragment();

			rChildren = replacement.childNodes;
			rLength = rChildren.length;
			for ( i=0; i<rLength; i++ ) {
				replacementFragment.appendChild( rChildren[i].cloneNode(true) );
			}

			target = mozileTarget.document.getElementsByTagName("body")[0];
			while ( target.hasChildNodes() ) { //remove all child nodes from body.
				target.removeChild( target.firstChild );
			}
			target.appendChild( replacementFragment );

		} catch (e) {
			mozileDebug(f,0,"Exception! "+e);
		}
	}
	else {
		mozileDebug(f,3,"XML context");
		//Note: cloning and replacing the whole documentElement locks the browser scrollbars.
		//Using a document fragment seems to cure this.
		try {
			replacementFragment = mozileTarget.document.createDocumentFragment();

			rChildren = replacement.childNodes;
			rLength = rChildren.length;
			for ( i=0; i<rLength; i++ ) {
				replacementFragment.appendChild( rChildren[i].cloneNode(true) );
			}

			target = mozileTarget.document.documentElement;
			while ( target.hasChildNodes() ) { //remove all child nodes from body.
				target.removeChild( target.firstChild );
			}
			target.appendChild( replacementFragment );

		} catch (e) {
			mozileDebug(f,0,"Exception! "+e);
		}

	}
}

// Dump Undo Stack
// Mostly for debugging (MAX...needs updating if you want to use it.)
/*function dumpUndoStack() {
	var out="Current: "+mozileTarget.undoCounter+"\n";
	for(var i=0;i<mozileTarget.undoStack.length;i++) {
		cur=mozileTarget.undoStack[i];
		out+="["+i+"] "+cur['name'] +": "+ cur['node'].tagName +"\n";
	}
	alert(out);
}*/ // end dumpUndoStack()

